module quicksort.

type quicksort (list int) -> (list int) -> o.
type pivot (list int) -> int -> o.
type partition int -> (list int) -> (list int) -> (list int) -> o.
% The pivot is the first element not in order.
% no pivot indicates list that's already sorted.
pivot [A,B|T] B :- A > B, !. 
pivot [A,B|T] P :- pivot [B|T] P.

% partition P A B C partitions A into B and C, based on pivot P
partition P [] [] [].
partition P [A|As] [A|Bs] Cs :- A =< P, partition P As Bs Cs.
partition P [A|As] Bs [A|Cs] :- A > P, partition P As Bs Cs.

quicksort L M :- pivot L P, !, partition P L A B,
  quicksort A As, quicksort B Bs, append As Bs M.
quicksort L L.

type append (list A) -> (list A) -> (list A) -> o.
append [] L L.
append [H|T] L [H|M] :- append T L M.

type timediff int -> int -> int -> int -> int -> o.
timediff S1 M1 S2 M2 D :- 
  MM1 is (M1 div 1000), MM2 is (M2 div 1000),
  SD is (S2 - S1),  SDM is (SD * 1000),
  MA2 is (MM2 + SDM), D is (MA2 - MM1).

type genlist int -> (list int) -> o.

genlist 0 [].
genlist N [A|As] :- time X Y, A is (Y mod 100), N2 is (N - 1),
  genlist N2 As.


test N :-
  genlist N L,
  time X0 Y0, time X1 Y1,   % second time more accurate
  quicksort L M,
  time X2 Y2,
  timediff X1 Y1 X2 Y2 D,   % d in milliseconds
  printterm std_out D, print"ms elapsed time\n".



