% Last update: 7/9/95

module parser.
import objsign lists.

type assoc  fo -> fo -> o.

type class  string -> list tm -> list string  -> list fo -> o.
doc  class  "Class    Vars       Private         Methods".

type subclass string -> string -> list tm -> list string -> list fo -> o.
doc  subclass "Super    Class     Vars       Private         Methods".

type parse_state_dec list tm ->  list fo -> list string -> list fo -> fo -> o. 
type parse_private_dec  list string -> list fo -> list fo -> fo -> o.

type parse_class   o -> o -> o.    
type parse_meth_list list fo -> list fo -> o.
type parse_structure fo -> fo -> o.

type parse_parameters list fo -> list fo -> o.
type parse_updates    list fo -> list fo -> o.
type parse_ops        list fo -> list fo -> o.

parse_class (class Name Vars Dec Meth) (classdef Name Def) :-
 pi id \  
  assoc (iref self) (iref id) =>  
  parse_state_dec Vars nil Dec Meth (Def id).

parse_class (subclass Super Name Vars Dec Meth) (subclassdef Super Name Def) :-
 pi id \  
  assoc (iref self) (iref id) =>
  parse_state_dec Vars nil Dec Meth (Def id).

parse_state_dec ((str V)::InVars)  OutVars  Dec Meth (fa_str Def) :-
 pi x \  
  assoc (istr V)  (istr x)  =>
  parse_state_dec InVars ((var x (istr " "))::OutVars) Dec Meth (Def x).

parse_state_dec ((num V)::InVars)  OutVars  Dec Meth (fa_str Def) :-
 pi x \  
  assoc (istr V)  (istr x)  =>
  parse_state_dec InVars ((var x (iint 0))::OutVars) Dec Meth (Def x).

parse_state_dec ((rf V)::InVars)  OutVars  Dec Meth (fa_str Def) :-
 pi x \  
  assoc (istr V) (istr x)  =>
  parse_state_dec InVars ((var x (iref null))::OutVars) Dec Meth (Def x).

parse_state_dec ((flag V)::InVars) OutVars Dec Meth (fa_str Def) :-
 pi x \
  assoc (istr V) (istr x) =>
  parse_state_dec InVars ((var x (ibool f))::OutVars) Dec Meth (Def x).

parse_state_dec     nil  Vars Dec  Meth Def:-
 parse_private_dec  Dec  Vars Meth Def.
   
parse_private_dec   nil Vars Meth (obj Id Meth_Def):-
 assoc (iref self) (iref Id),  
 parse_meth_list Meth Def,
 append Vars Def Meth_Def.

parse_private_dec (MethName::Others) Vars Meth (fa_str Def) :-
 pi x \  
  assoc (istr MethName) (istr x) => 
  parse_private_dec Others Vars Meth (Def x).

parse_meth_list nil nil.

parse_meth_list ((meth Name Str)::Ms) ((meth X Str1)::Ds) :-
 assoc (istr Name) (istr X),!,
 parse_structure Str Str1,
 parse_meth_list Ms Ds.
  
parse_meth_list ((meth Name Str)::Ms) ((meth Name Str1)::Ds) :-
 parse_structure Str Str1,
 parse_meth_list Ms Ds.

parse_structure (fa_fo Body) (fa_fo Body1):-
 pi x \ parse_structure (Body x) (Body1 x).
  
parse_structure (fa_str Body) (fa_str Body1):-
 pi x \ parse_structure (Body x) (Body1 x). 
 
parse_structure (fa_int Body) (fa_int Body1):-
 pi x \ parse_structure (Body x) (Body1 x).
  
parse_structure (fa_ref Body) (fa_ref Body1):-
 pi x \ parse_structure (Body x) (Body1 x).

parse_structure (fa_bool Body) (fa_bool Body1):-
 pi x \ parse_structure (Body x) (Body1 x).

parse_structure (C & D) (C1 & D1):-
 parse_structure C C1,
 parse_structure D D1.
 
parse_structure (pars_updates_body Pars Updates Ops) 
                (pars_updates_body Pars1 Updates1 Ops1):-
 parse_parameters Pars Pars1,
 parse_updates Updates Updates1,
 parse_ops Ops Ops1.

% parse_structure Value Value.
%parse_structure (istr Value) (istr Value).
%parse_structure (iint Value) (iint Value).
%parse_structure (iref Value) (iref Value).
 
parse_structure  Str Str.

parse_parameters ((iref self)::Others) ((iref Id)::Others1) :-
 assoc (iref self) (iref Id),
 !,parse_parameters Others Others1.
 
parse_parameters (P::Others) (P::Others1) :-
 parse_parameters Others Others1.
 
parse_parameters nil nil.
  
parse_updates ((get Name Body)::Updates) ((get X Body1)::Updates1) :-
 assoc (istr Name) (istr X),
 !,parse_structure Body Body1,
 parse_updates Updates Updates1.

parse_updates ((put Name Body)::Updates) ((put X Body1)::Updates1) :-
 assoc (istr Name) (istr X),
 !,parse_structure Body Body1,
 parse_updates Updates Updates1.

parse_updates ((get Name Body)::Updates) ((get Name Body1)::Updates1) :-
 parse_structure Body Body1,
 parse_updates Updates Updates1.

parse_updates ((put Name Body)::Updates) ((put Name Body1)::Updates1) :-
 parse_structure Body Body1,
 parse_updates Updates Updates1.

parse_updates ((addmeth Name Body)::Updates) ((addmeth Name Body1)::Updates1) :-
 parse_structure Body Body1,
 parse_updates Updates Updates1.

parse_updates ((killmeth Name)::Updates) ((killmeth Name)::Updates1) :-
 parse_updates Updates Updates1.
 
parse_updates nil nil.

parse_ops (local Ops) (local Ops1):-
 pi x \ parse_ops (Ops x) (Ops1 x).

parse_ops ((call self Name Pars)::Ops) ((call Id X Pars1)::Ops1) :-
 assoc (iref self) (iref Id),
 assoc (istr Name) (istr X),
 !,parse_parameters Pars Pars1,
 parse_ops Ops Ops1.

parse_ops ((call self Name Pars)::Ops) ((call Id Name Pars1)::Ops1) :-
 assoc (iref self) (iref Id),
 !,parse_parameters Pars Pars1,
 parse_ops Ops Ops1.

parse_ops (Op::Others) (Op::Others1) :-
 parse_ops Others Others1.

parse_ops nil nil.













