module randomlams.

% assume random R X. will get a random number between 0 and R-1.
% 0 - free variable
% 1 - bound variable
% 2 - app term
% 3 - abs term


% free var
genterm 0 BV (freevar X).

% bound var
genterm 1 [] (freevar V) :- !.   % defaults to free variable.
genterm 1 BV V :-
  length BV L, random L N,
  nthmemb N BV V.

% app term 
genterm 2 BV (app A B) :- 
  random 4 N, random 4 M,
  genterm N BV A, genterm M BV B.

% abs term
genterm 3 BV (abs A) :- 
  random 4 N,
  (pi v\ (genterm N (v::BV) (A v))).


% object-level head normalization
%hnorm A B :- printterm std_out (hnorm A B), print " <-\n", fail.
%hnormw A B :- printterm std_out (hnormw A B), print " <-\n", fail.

hnorm (freevar V) (freevar V).
hnorm X Y :- bvar X, X = Y.
%hnorm X X :- bvar X.
hnorm (abs A) (abs B) :- pi v\ (bvar v => hnorm (A v) (B v)).
hnorm (app A B) C :- hnormw A D,  
    ((D = (abs D2), hnorm (D2 B) C); 
     (app D B) = C), !.

%local hnormaux, hnormwaux i -> i -> i -> o.
%hnormaux (abs D2) B C :- !, hnorm (D2 B) C.
%hnormaux D B (app D B).

hnormw (freevar V) (freevar V).
hnormw X Y :- bvar X, X = Y.
%hnormw X X :- bvar X.
hnormw (abs A) (abs A).
hnormw (app A B) C :- hnormw A D, 
    ((D = (abs D2), hnormw (D2 B) C); 
     (app D B) = C), !.

% (abs x\ (app (app (abs y\ (abs z\ y)) (app x x)) x)). 
%hnormwaux (abs D2) B C :- !, hnormw (D2 B) C.
%hnormwaux D B (app D B).





% generate lambda terms 
% (go N) generates N random lambda terms

go Max:- 
  open_socket "127.0.0.1" 1917 In Out,
    (instream In => (outstream Out => go2 Max)),
  output Out "bye\n", 
  close_in In, close_out Out.


% this algorithm tries to cut down on the number of single freevar terms:
go2 0.
go2 M :- M > 0, random 2 N, N2 is (N + 2), genterm N2 [] T, 
   print "lambdaterm (", printterm std_out T, print ").\n", 
   M2 is (M - 1), go2 M2.
      
randterm T :-
  open_socket "127.0.0.1" 1917 In Out, 
  (instream In => (outstream Out => 
	( random 6 N, genterm N [] T))),
  output Out "bye\n", flush Out,
  close_in In, close_out Out.


random Range Num :-
  instream In, outstream Out,
  RS is (int_to_string Range),
  output Out RS, output Out "\n", flush Out,
  input_line In Ins,
  open_string Ins STIN,
  readterm STIN Num, 
  close_in STIN.

%  Rand is (string_to_int Ins), Num is (Rand - 48), Num < 10, Num >= 0.
% The more general way (the above option only gets one-digit values):
%  open_string Ins STIN,
%  readterm STIN Num, 
%  print "Got number: ", printterm std_out Num, print "\n",
%  close_in STIN. 



%  ---  utilities.    

length [] 0.
length [H|T] N :- length T M, N is (M + 1).

nthmemb 0 [H|T] H.
nthmemb N [H|T] U :- N2 is (N - 1), nthmemb N2 T U.


